/*
    libporty - unified pointer handling
    Copyright (C) 2011..2012  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page: http://repo.hu/projects/libporty
    Contact (email and IRC): http://igor2.repo.hu/contact.html
*/

#include <stdlib.h>
#include "libporty/config.h"
#include "pointers.h"

static const char *hdigit="0123456789ABCDEF";

static P_inline void c2hex(unsigned char **out, unsigned char in)
{
	int nibble;
	nibble = (in & 0xF0) >> 4;
	**out = hdigit[nibble];
	(*out)++;
	nibble = (in & 0x0F);
	**out = hdigit[nibble];
	(*out)++;
}

int P_ptr_print(char *str, void *ptr)
{
	unsigned char *in, *out, *start, *end;
/* we could simply use an int wide enough for holding a ptr to simplify the
   code blow. However, finding out what int is sufficient on a given platform
   is probably more expensive than this hack.*/

	start = end = (unsigned char *)&ptr;
	end += sizeof(void *) - 1;
	out = (unsigned char *)str;
	*out = 'P'; out++;
	*out = 'x'; out++;
#ifdef P_HOST_LSB
	for(in = end; in >= start; in--)
		c2hex(&out, *in);
#else
#	ifdef P_HOST_MSB
	for(in = start; in <= end; in++)
		c2hex(&out, *in);
#	else
#		error Host is neither LSB or MSB? - configuration must be broken
#	endif
#endif
	*out = '\0';
	return out - (unsigned char *)str;
}

char *P_ptr_aprint(void *ptr)
{
	char *str;

	str = malloc(P_PTR_STRWIDTH);
	P_ptr_print(str, ptr);
	return str;
}

static P_inline int hex2c(const unsigned char *str, unsigned char *out)
{
	int nibble;
	nibble = *str - '0';
	if (nibble > 9)
		nibble = *str - 'A' + 10;
	if ((nibble > 16) || (nibble < 0))
		return -1;
	*out = nibble << 4;

	str++;
	nibble = *str - '0';
	if (nibble > 9)
		nibble = *str - 'A' + 10;
	if ((nibble > 16) || (nibble < 0))
		return -1;

	*out |= nibble;
	return 0;
}

int P_ptr_scan(const char *str, void **ptr)
{
	unsigned char *out;
	const unsigned char *in, *end;
	if ((str[0] != 'P') || (str[1] != 'x'))
		return -1;
	end = out = (unsigned char *)ptr;
#ifdef P_HOST_LSB
	out += sizeof(void *)-1;

	for(in = (const unsigned char *)str + 2; *in != '\0'; in += 2, out--) {
		if (out < end)
			return -1;
		if (hex2c(in, out) != 0)
			return -1;
	}
	if (out+1 != end)
		return -1;
#else
#	ifdef P_HOST_MSB
	end += sizeof(void *);

	for(in = (const unsigned char *)str + 2; *in != '\0'; in += 2, out++) {
		if (out >= end)
			return -1;
		if (hex2c(in, out) != 0)
			return -1;
	}
	if (out != end)
		return -1;
#	endif
#endif
	return 0;
}

