/*
 * Copyright (c) 2003 Anders Magnusson (ragge@ludd.luth.se).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Copyright(C) Caldera International Inc. 2001-2002. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * Redistributions of source code and documentation must retain the above
 * copyright notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * All advertising materials mentioning features or use of this software
 * must display the following acknowledgement:
 *         This product includes software developed or owned by Caldera
 *        International, Inc.
 * Neither the name of Caldera International, Inc. nor the names of other
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * USE OF THE SOFTWARE PROVIDED FOR UNDER THIS LICENSE BY CALDERA
 * INTERNATIONAL, INC. AND CONTRIBUTORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL CALDERA INTERNATIONAL, INC. BE LIABLE
 * FOR ANY DIRECT, INDIRECT INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OFLIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

 /***************************************************************************
    This is a modified verison for c99tree; these modifications are
    Copyright (c) 2010-2015 Tibor Palinkas.
    Original file version marking:
    cgram.y,v 1.296 2010/08/11 14:08:44 ragge Exp $
    Modifications are licensed the same way as the original file.
 *****************************************************************************/

/*
 * Comments for this grammar file. Ragge 021123
 *
 * ANSI support required rewrite of the function header and declaration
 * rules almost totally.
 *
 * The lex/yacc shared keywords are now split from the keywords used
 * in the rest of the compiler, to simplify use of other frontends.
 */

/*
 * At last count, there were 3 shift/reduce and no reduce/reduce conflicts
 * Two was funct_idn and the third was "dangling else".
 * (currently we have some more -- Igor2)
 */

/*%locations*/
%error-verbose


/*
 * Token used in C lex/yacc communications.
 */
%token        C_STRING        /* a string constant */
%token        C_ICON          /* an integer constant */
%token        C_FCON          /* a floating point constant */
%token        C_NAME          /* an identifier */
%token        C_TYPENAME      /* a typedef'd name */
%token        C_ANDAND        /* && */
%token        C_OROR          /* || */
%token        C_GOTO          /* unconditional goto */
%token        C_RETURN        /* return from function */
%token        C_TYPE          /* a type */
%token        C_TYPESPEC      /* a type specifier (signed/unsigned) */
%token        C_CLASS         /* a storage class */
%token        C_TYPEDEF       /* typedef command */
%token        C_ASOP          /* assignment ops */
%token        C_RELOP         /* <=, <, >=, > */
%token        C_EQUOP         /* ==, != */
%token        C_DIVOP         /* /, % */
%token        C_SHIFTOP       /* <<, >> */
%token        C_INCOP         /* ++, -- */
%token        C_UNOP          /* !, ~ */
%token        C_STROP         /* ., -> */
%token        C_STRUCT
%token        C_IF
%token        C_ELSE
%token        C_SWITCH
%token        C_BREAK
%token        C_CONTINUE
%token        C_WHILE
%token        C_DO
%token        C_FOR
%token        C_DEFAULT
%token        C_CASE
%token        C_SIZEOF
%token        C_ALIGNOF
%token        C_ENUM
%token        C_ELLIPSIS
%token        C_QUALIFIER
%token        C_INLINE
%token        C_ASM
%token        NOMATCH
%token        C_TYPEOF        /* COMPAT_GCC */
%token        C_ATTRIBUTE     /* COMPAT_GCC */
%token        PCC_OFFSETOF
%token        C_PRAGMA
%token        C_CUSTOP_B        /* custom binary operators */

/*
 * Precedence
 */
%left ','
%right '=' C_ASOP
%right '?' ':'
%left C_OROR
%left C_ANDAND
%left '|'
%left '^'
%left '&'
%left C_EQUOP
%left C_RELOP
%left C_SHIFTOP
%left '+' '-'
%left '*' C_DIVOP
%right C_UNOP  CUSTOP_U
%right C_INCOP C_SIZEOF
%left '[' '(' C_STROP C_CUSTOP_B
%{
# define YLINEDEF  __LINE__
# include "c99tree-cc.h"
# include "scope.h"
# include <stdarg.h>
# include <string.h>
# include <stdlib.h>

static const char *extract_id_from_cast(c99tree_cc_t *ctx, c99c_node_t *n);
static int is_designator(c99c_node_t *n);

/* use the location of node n or invalid location if n is NULL */
#define C99NULL_LOC_OR_INV(n) ((n == NULL) ? &ctx->loc_invalid : NULL)

/* different versions of bison has different conventions of passing multiple args
   to lex while lex forces its own idea. This wrapper gets only what bison
   is willing to pass and calls lex how lex wants it */
static int yylex_wrap(void *s, c99tree_cc_t *ctx)
{
	return c99c_lex(s, ctx->scanner, ctx);
}

#define YYLEX_PARAM ctx
#undef yylex
#define yylex yylex_wrap
%}

%pure-parser
%parse-param {c99tree_cc_t *ctx}
%lex-param {c99tree_cc_t *ctx}


%union {
	c99c_node_t *node;
	c99c_strloc_t string;
}

/* define types */
%start ext_def_list

%type <string> ';' '{' '}' ',' ':' '=' '(' ')' '[' ']' '&' '-' '+' '*' '^'
%type <string> '|' '?' C_SIZEOF C_ALIGNOF

%type <node>    external_def kr_args declaration asmstatement arg_dcl_list
                arg_declaration block_item_list block_item compoundstmt
                init_declarator init_declarator_list init_list label moe
                moe_list asm_mvol statement struct_dcl_list struct_declaration
                struct_declarator struct_declarator_list switch_label
                xnfdeclarator_typename init_declarator_typename

%type <node>    ifelprefix ifprefix whprefix forprefix doprefix switchpart

%type <node>    e .e ae be term enum_dcl struct_dcl cast_type
                declarator declarator_typename elist type_sq cf_spec
                merge_attribs
                parameter_declaration abstract_declarator initializer
                parameter_type_list parameter_list addrlbl
                declaration_specifiers designation type_sq_list
                identifier_list arg_param_list type_qualifier_list
                designator_list designator xasm oplist oper cnstr funtype
                typeof attribute attribute_specifier
                attribute_list attr_spec_list attr_var_pre attr_var_post preproc
                abstract_declarator_ end  endc virt_statement

%type <node>    string string_
%type <string>  C_STRING
%type <node>    str_head
%type <node>    xnfdeclarator enum_head

%type <string>  C_STRUCT C_RELOP C_DIVOP C_SHIFTOP
                C_ANDAND C_OROR C_STROP C_INCOP C_UNOP C_ASOP C_EQUOP C_CASE
                C_CUSTOP_B

%type <string>  C_TYPESPEC C_TYPE C_QUALIFIER C_ICON C_FCON C_CLASS C_TYPEDEF
%type <node>    C_TYPENAME
%type <string>  C_NAME C_CONTINUE C_ELLIPSIS
%type <string>  C_PRAGMA C_FOR C_WHILE C_GOTO C_RETURN C_TYPEOF
%type <string>  str_name C_SWITCH C_IF C_DO C_INLINE C_BREAK C_DEFAULT


%type <string>   opt_extra_semi C_ATTRIBUTE C_ENUM
%%

/* entry point - a list of external defs */
ext_def_list:
		  ext_def_list external_def         { if ($2 != NULL) {if (ctx->AST == NULL) ctx->AST = $2; else C99Nappend(ctx, ctx->AST, $2);} }
		| /* empty */
		;

/* source is made of declaration, function definition, preproc, or asm */
external_def:
		  funtype kr_args compoundstmt {
			$$ = C99Nalloc_wloc(ctx, T_FUNCDEF, c99c_strclone($1->next->name), $1, &$1->loc, &$3->loc);
			C99Nloc_cpy_start(ctx, $$, $1);
			if (($1->next->children->name != NULL) && (strcmp($1->next->children->name, "K&R/1") == 0))
				C99Nchappend(ctx, $1->next, C99Nalloc(ctx, T_ARGLIST, c99c_strclone("K&R/2"), $2));
			C99Nchappend(ctx, $$, $3);
		}
		| declaration        { $$ = $1; }
		| asmstatement ';'   { $$ = C99Nalloc(ctx, T_FUNCDEF, NULL, $1); }
		| preproc            { $$ = $1; }
		| ';'                { $$ = NULL; }
		| error              { $$ = NULL; }
		;

/* function declaration */
funtype:
		  /* no type given */ declarator {
			$$ = C99Nalloc(ctx, T_TYPE, NULL, NULL);
			C99Nappend(ctx, $$, $1);
		}
		| declaration_specifiers declarator {
			$$ = $1;
			C99Nappend(ctx, $1, $2);
		}
		;

kr_args:
		  /* empty */       { $$ = NULL; }
		| arg_dcl_list      { $$ = $1;   }
		;

declaration_specifiers:
		  merge_attribs    { $$ = C99Nalloc_wloc(ctx, T_TYPE, NULL, $1, &$1->loc, NULL); }
		;

merge_attribs:
		  type_sq                 { $$ = $1; }
		| merge_attribs type_sq   { $$ = C99Nappend(ctx, $1, $2); }
		| cf_spec                 { $$ = $1; }
		| merge_attribs cf_spec   { $$ = C99Nappend(ctx, $1, $2); }
		;

type_sq:
		  C_TYPE                { $$ = C99Nalloc_tok(ctx, T_TYPENAME, &($1), NULL); }
		| C_TYPESPEC            { $$ = C99Nalloc_tok(ctx, T_TYPESPEC, &($1), NULL); }
		| C_TYPENAME            { $$ = $1; free(ctx->c_typename); ctx->c_typename = NULL; }
		| struct_dcl            { $$ = $1; }
		| enum_dcl              { $$ = $1; }
		| C_QUALIFIER           { $$ = C99Nalloc_tok(ctx, T_QUALIFIER, &($1), NULL);  }
		| typeof                { $$ = $1; }
		| attribute_specifier   { $$ = $1; }
		;

cf_spec:
		  C_CLASS       { $$ = C99Nalloc_tok(ctx, T_CLASS, &($1), NULL); }
		| C_INLINE      { $$ = C99Nalloc_tok(ctx, T_CLASS, &($1), NULL); }
		;

typeof:
		  C_TYPEOF '(' e ')'            { $$ = C99Nalloc(ctx, T_TYPEOF, NULL, $3); c99c_loc_cpy_start(&$$->loc, &$1.loc); }
		| C_TYPEOF '(' cast_type ')'    { $$ = C99Nalloc(ctx, T_TYPEOF, NULL, $3); c99c_loc_cpy_start(&$$->loc, &$1.loc); }

		;

attribute_specifier:
		  C_ATTRIBUTE '(' '(' attribute_list ')' ')'   { $$ = C99Nalloc_wloc(ctx, T_ATTRLIST, NULL, $4, &$1.loc, &$6.loc); }
		;

attribute_list:
		  attribute                       { $$ = $1; }
		| attribute_list ',' attribute    { $$ = C99Nappend(ctx, $1, $3); }
		;

attribute:
		  /* EMPTY */               { $$ = NULL; }
		| C_NAME                    { $$ = C99Nalloc_tok(ctx, T_ATTR, &($1), NULL); }
		| C_NAME '(' elist ')'      { $$ = C99Nalloc_tok(ctx, T_ATTR, &($1), $3); }
		| C_QUALIFIER               { $$ = C99Nalloc_tok(ctx, T_ATTR, &($1), NULL); }
		;

declarator:
		  '*' declarator                            { $$ = $2; C99NTYinsert(ctx, $$, C99Nalloc_tok(ctx, T_PTR, &$1, NULL)); }
		| '*' type_qualifier_list declarator        { $$ = $3; C99NTYinsert(ctx, $$, $2); C99Nchinsert(ctx, $$->children, C99Nalloc_tok(ctx, T_PTR, &$1, NULL));}
		| C_NAME                                    { $$ = C99Nalloc_tok(ctx, T_ID, &($1), NULL);}
		| '(' attr_spec_list declarator ')'         { $$ = $3; if ($3->children != NULL) C99Nchinsert(ctx, $3->children, $2); else C99Nchinsert(ctx, $3, $2); }
		| '(' declarator ')'                        { $$ = $2; }
		| declarator '[' ae ']'                     { $$ = $1; C99NTYinsert(ctx, $$, C99Nalloc_wloc(ctx, T_INDEX, NULL, $3, &$2.loc, &$4.loc));  }
		| declarator '[' merge_attribs  ae ']'      { $$ = $1; C99NTYinsert(ctx, $$, C99Nalloc_wloc(ctx, T_INDEX, NULL, $3, &$2.loc, &$5.loc)); C99Nappend(ctx, $3, $4); }
		| declarator '[' merge_attribs ']'          { $$ = $1; C99NTYinsert(ctx, $$, C99Nalloc_wloc(ctx, T_INDEX, NULL, $3, &$2.loc, &$4.loc)); }
		| declarator '[' ']'                        { $$ = $1; C99NTYinsert(ctx, $$, C99Nalloc_wloc(ctx, T_INDEX, NULL, NULL, &$2.loc, &$3.loc)); }
		| declarator '[' '*' ']'                    { c99c_node_t *n; $$ = $1; n = C99Nalloc_wloc(ctx, T_ID, c99c_strclone("*"), NULL, &$3.loc, &$3.loc);         C99NTYinsert(ctx, $$, C99Nalloc_wloc(ctx, T_INDEX, c99c_strclone("abstract_*"), n, &$2.loc, &$4.loc)); }
		| declarator '(' parameter_type_list ')'    { $$ = $1; C99NTYinsert(ctx, $$, C99Nalloc_wloc(ctx, T_ARGLIST, NULL, $3, &$2.loc, &$4.loc)); }
		| declarator '(' identifier_list ')'        { $$ = $1; C99NTYinsert(ctx, $$, C99Nalloc_wloc(ctx, T_ARGLIST, c99c_strclone("K&R/1"), $3, &$2.loc, &$4.loc)); }
		| declarator '(' ')'                        { $$ = $1; C99NTYinsert(ctx, $$, C99Nalloc_wloc(ctx, T_ARGLIST, NULL, NULL, &$2.loc, &$3.loc)); }
		;

declarator_typename:
		  '*' declarator_typename                            { $$ = $2; C99NTYinsert(ctx, $$, C99Nalloc_tok(ctx, T_PTR, &$1, NULL)); }
		| '*' type_qualifier_list declarator_typename        { $$ = $3; C99NTYinsert(ctx, $$, $2); C99Nchinsert(ctx, $$, C99Nalloc_tok(ctx, T_PTR, &$1, NULL));}
		| C_TYPENAME                                         { $$ = C99Nalloc(ctx, T_ID, c99c_strclone($1->name), NULL); c99c_scope_negative_typedef_add(ctx, $1->name); C99Nloc_cpy_start(ctx, $$, $1); C99Nfree(ctx, $1); free(ctx->c_typename); ctx->c_typename = NULL; }
		| '(' attr_spec_list declarator_typename ')'         { $$ = C99Nappend(ctx, $2, $3); }
		| '(' declarator_typename ')'                        { $$ = $2; }
		| declarator_typename '[' ae ']'                     { $$ = $1; C99NTYinsert(ctx, $$, C99Nalloc_wloc(ctx, T_INDEX, NULL, $3, &$2.loc, &$4.loc)); }
		| declarator_typename '[' ']'                        { $$ = $1; C99NTYinsert(ctx, $$, C99Nalloc_wloc(ctx, T_INDEX, NULL, NULL, &$2.loc, &$3.loc)); }
		| declarator_typename '[' '*' ']'                    { $$ = $1; C99NTYinsert(ctx, $$, C99Nalloc_wloc(ctx, T_INDEX, c99c_strclone("*"), NULL, &$2.loc, &$4.loc)); }
		| declarator_typename '(' parameter_type_list ')'    { $$ = $1; C99NTYinsert(ctx, $$, C99Nalloc_wloc(ctx, T_ARGLIST, NULL, $3, &$2.loc, &$4.loc)); }
		| declarator_typename '(' identifier_list ')'        { $$ = $1; C99NTYinsert(ctx, $$, C99Nalloc_wloc(ctx, T_ARGLIST, c99c_strclone("K&R/1"), $3, &$2.loc, &$4.loc)); }
		| declarator_typename '(' ')'                        { $$ = $1; C99NTYinsert(ctx, $$, C99Nalloc_wloc(ctx, T_ARGLIST, NULL, NULL, &$2.loc, &$3.loc)); }
		;

/* add qualifiers to a declaration */
type_qualifier_list:
		  C_QUALIFIER                                  { $$ = C99Nalloc_tok(ctx, T_QUALIFIER, &($1), NULL); }
		| type_qualifier_list C_QUALIFIER              { $$ = C99Nappend(ctx, $1, C99Nalloc_tok(ctx, T_QUALIFIER, &($2), NULL)); }
		| attribute_specifier                          { $$ = $1; }
		| type_qualifier_list attribute_specifier      { $$ = C99Nappend(ctx, $1, $2); $2->name = c99c_strclone("pre"); }
		;

/* when a list of IDs is given without types - currently used only in K&R first set of arguments */
identifier_list:
		  C_NAME                          { $$ = C99Nalloc(ctx, T_ARG, NULL, C99Nalloc_tok(ctx, T_ID, &($1), NULL)); }
		| identifier_list ',' C_NAME      { $$ = C99Nappend(ctx, $1, C99Nalloc(ctx, T_ARG, NULL, C99Nalloc_tok(ctx, T_ID, &($3), NULL))); C99Nloc_cpy_start(ctx, $$, $1); }
		;

parameter_type_list:
		  parameter_list                    { $$ = $1; }
		| parameter_list ',' C_ELLIPSIS     { $$ = C99Nappend(ctx, $1, C99Nalloc(ctx, T_ARG, NULL, C99Nalloc(ctx, T_ELLIPSIS, NULL, NULL)));  C99Nloc_cpy_start(ctx, $$, $1); }
		;

parameter_list:
		  parameter_declaration                       { $$ = C99Nalloc(ctx, T_ARG, NULL, $1); }
		| parameter_list ',' parameter_declaration    { c99c_node_t *n = C99Nalloc(ctx, T_ARG, NULL, $3); $$ = C99Nappend(ctx, $1, n); }
		;

parameter_declaration:
		  declaration_specifiers declarator attr_var_post {
			$$ = $1;
			C99Nappend(ctx, $$, $2);
			C99Nappend(ctx, $$, $3);
		}
		| declaration_specifiers abstract_declarator    { $$ = C99Nappend(ctx, $1, $2); }
		| declaration_specifiers                        { $$ = $1; }
		;

/* anonymous declarations */
abstract_declarator:
	abstract_declarator_ { $$ = C99Nalloc_wloc(ctx, T_ID, c99c_strclone("<anonymous>"), $1, &ctx->loc_invalid, &ctx->loc_invalid); }
	;

abstract_declarator_:
		  '*'                             { $$ = C99Nalloc_tok(ctx, T_PTR, &$1, NULL); }
		| '*' type_qualifier_list         { $$ = C99Nappend(ctx, C99Nalloc_tok(ctx, T_PTR, &$1, NULL), $2); }
		| '*' abstract_declarator_        { $$ = C99Nappend(ctx, C99Nalloc_tok(ctx, T_PTR, &$1, NULL), $2); }
		| '*' type_qualifier_list abstract_declarator_ {
			$$ = $3;
			C99Nappend(ctx, $$, $2);
			C99Nappend(ctx, $$, C99Nalloc_tok(ctx, T_PTR, &$1, NULL));
		}
		| '(' abstract_declarator_ ')'   { $$ = $2; }
		| '[' ']' attr_var_post          { $$ = C99Nappend(ctx, C99Nalloc_wloc(ctx, T_INDEX, c99c_strclone("abstract_1"), NULL, &$1.loc, &$2.loc), $3); }
		| '[' ae ']' attr_var_post {
			$$ = C99Nalloc_wloc(ctx, T_INDEX, c99c_strclone("abstract_2"), NULL, &$1.loc, &$3.loc);
			C99Nchinsert(ctx, $$, $2);
			C99Nappend(ctx, $$, $4);
		}
		| abstract_declarator_ '[' ']' attr_var_post {
			if ($4 != NULL) {
				$$ = $4;
				C99Nappend(ctx, $$, C99Nalloc_wloc(ctx, T_INDEX, c99c_strclone("abstract_3a"), NULL, &$2.loc, &$3.loc));
			}
			else
				$$ = C99Nalloc_wloc(ctx, T_INDEX, c99c_strclone("abstract_3b"), NULL, &$2.loc, &$3.loc);
			C99Nappend(ctx, $$, $1);
		}
		| abstract_declarator_ '[' ae ']' attr_var_post {
			$$ = C99Nalloc_wloc(ctx, T_INDEX, c99c_strclone("abstract_4"), $3, &$2.loc, &$4.loc);
			C99Nappend(ctx, $$, $5);
			C99Nappend(ctx, $$, $1);
		}
		| '(' ')'                           { $$ = C99Nalloc_wloc(ctx, T_ARGLIST, NULL, NULL, &$1.loc, &$2.loc); }
		| '(' parameter_type_list ')'       { $$ = C99Nalloc_wloc(ctx, T_ARGLIST, NULL, $2, &$1.loc, &$3.loc); }
		|  abstract_declarator_ '(' ')'     { c99c_node_t *t; $$ = C99Nappend(ctx, t=C99Nalloc(ctx, T_TYPE, NULL, C99Nalloc_wloc(ctx, T_ARGLIST, NULL, NULL, &$2.loc, &$3.loc)), NULL); C99Nchappend(ctx, t, $1); }
		|  abstract_declarator_ '(' parameter_type_list ')' { c99c_node_t *t; $$ = C99Nappend(ctx, t=C99Nalloc(ctx, T_TYPE, NULL, C99Nalloc_wloc(ctx, T_ARGLIST, NULL, $3, &$2.loc, &$4.loc)), NULL); C99Nchappend(ctx, t, $1); }
		;

/*
 * K&R arg declaration, between ) and {
 */
arg_dcl_list:
		  arg_declaration                   { $$ = $1; }
		| arg_dcl_list arg_declaration      { $$ = C99Nappend(ctx, $1, $2); }
		;


arg_declaration:
		declaration_specifiers arg_param_list ';'  { $$ = C99Nalloc(ctx, T_ARG, NULL, C99Nappend(ctx, $1, $2)); }
		;

arg_param_list:
		  declarator attr_var_post                 { $$ = $1; C99Nchappend(ctx, $$, $2); }
		| arg_param_list ',' declarator attr_var_post {
			$$ = $1;
			C99Nappend(ctx, $$, $3);
			C99Nchappend(ctx, $3, $4);
		}
		;

/*
 * Declarations in beginning of blocks.
 */
block_item_list:
		  block_item                       { $$ = $1; }
		| block_item_list block_item       { if ($1 == NULL) $$ = $2; else $$ = C99Nappend(ctx, $1, $2); }
		;

block_item:
		  declaration       { $$ = $1; }
		| statement         { $$ = $1; }
		;

/* Variable declaration */
declaration:
		  declaration_specifiers ';' {
				c99c_node_t *idp;

				if (($1->children != NULL) && ($1->children->next != NULL)) {
					for(idp = $1->children; idp->next->next != NULL; idp = idp->next);
				}
				else
					idp = NULL;

				if ((idp != NULL) && (idp->next->type == T_TYPECUSTOM)) {
					/* workaround for typename redeclared as variable */
					const char *id;

					id = idp->next->name;
					$$ = C99Nalloc_wloc(ctx, T_DECL, NULL, C99Nalloc_wloc(ctx, T_ID, c99c_strclone(id), NULL, &idp->next->loc, &idp->next->loc), &$1->loc, &$2.loc);
					c99c_scope_negative_typedef_add(ctx, id);
					C99Nfree(ctx, idp->next);
					idp->next = NULL;
					C99Nchinsert(ctx, $$, $1);
				}
				else
					$$ = C99Nalloc_wloc(ctx, T_DECL, NULL, $1, &$1->loc, &$2.loc);
		}
		| C_TYPEDEF declaration_specifiers init_declarator_list ';' {
			 $$ = C99Nalloc_wloc(ctx, T_TYPEDEF, NULL, $2, &$1.loc, &$4.loc);
			 C99Nchappend(ctx, $$, $3);
			 c99c_scope_typedef_add(ctx, $2, $3);
		}
		| C_TYPEDEF declaration_specifiers ';' {
			/* an existing type redefined. Check if the type was originally defined
			   by -Y and overwrite it if so */
			const c99c_node_t *orig;
			c99c_node_t *last = NULL, *id = $2->children;
			while(id->next != NULL) {
				last = id;
				id = id->next;
			}

			/* check whether it is valid */
			orig = c99c_scope_get_typedef(ctx, id->name);
			if (orig == NULL) {
				ctx->error_hint = id->name;
				yyerror(ctx, "attemt to redefine builtin type");
			}
			else if (!orig->redefinable) {
				ctx->error_hint = id->name;
				yyerror(ctx, "attemt to redefine type");
			}
			if (last == NULL) {
				yyerror(ctx, "typedef requires at least two types");
			}

			/* all fine, redeclare the type */
			last->next = NULL;
			id->type = T_ID;
			$$ = C99Nalloc(ctx, T_TYPEDEF, NULL, $2);
			C99Nchappend(ctx, $$, id);
			c99c_scope_typedef_add(ctx, $2, id);
			free($1.str);
		}
		| declaration_specifiers init_declarator_list ';' {
			$$ = C99Nalloc_wloc(ctx, T_DECL, NULL, $1, &$1->loc, &$3.loc);
			C99Nappend(ctx, $1, $2);
		}
		;

/* Normal declaration of variables. */
init_declarator_list:
		  init_declarator { $$ = $1; }
		| init_declarator_list ','  attr_var_pre init_declarator_typename {
			$$ = $1;
			if ($3 != NULL) {
				C99Nchinsert(ctx, $4, $3);
				C99Nappend(ctx, $$, $4);
			}
			else
				C99Nappend(ctx, $$, $4);
		}
		| init_declarator_list ',' attr_var_pre init_declarator {
			$$ = $1;
			if ($3 != NULL) {
				C99Nchinsert(ctx, $4, $3);
				C99Nappend(ctx, $$, $4);
			}
			else
				C99Nappend(ctx, $$, $4);
		}
		;

enum_dcl:
		  enum_head '{' moe_list optcomma '}' {
			$$ = $1;
			C99Nchappend(ctx, $$, $3);
			c99c_loc_cpy_end(&$$->loc, &$5.loc);
		}
		| C_ENUM C_NAME    { $$ = C99Nalloc_tok(ctx, T_ENUM, &($2), NULL); }
		;

enum_head:
		  C_ENUM             { $$ = C99Nalloc_tok(ctx, T_ENUM, &($1), NULL); }
		| C_ENUM C_NAME      { $$ = C99Nalloc_tok(ctx, T_ENUM, &($2), NULL); }
		;

moe_list:
		  moe                { $$ = $1; }
		| moe_list ',' moe   { $$ = C99Nappend(ctx, $1, $3); }
		;

moe:
		  C_NAME             { $$ = C99Nalloc_tok(ctx, T_ID, &($1), NULL); }
		| C_TYPENAME         { $$ = C99Nalloc(ctx, T_ID, ctx->c_typename, NULL); ctx->c_typename = NULL; C99Nloc_cpy_start(ctx, $$, $1); C99Nfree(ctx, $1); }
		| C_NAME '=' be      { $$ = C99Nalloc_tok(ctx, T_ID, &($1), C99Nalloc(ctx, T_INIT, NULL, $3)); }
		| C_TYPENAME '=' be  {
			$$ = C99Nalloc(ctx, T_ID, ctx->c_typename, C99Nalloc(ctx, T_INIT, NULL, $3));
			C99Nloc_cpy_start(ctx, $$, $1);
			$$->loc.twin_parent = $1->loc.twin_parent;
			ctx->c_typename = NULL;
			C99Nfree(ctx, $1);
		}
		;

struct_dcl:
		  str_head '{' struct_dcl_list '}' {
			$$ = $1;
			C99Nchappend(ctx, $$, $3);
			c99c_loc_cpy_end(&$$->loc, &$4.loc);
		}
		| C_STRUCT attr_var_pre C_NAME {
			if (*($1.str) == 'u')
				$$ = C99Nalloc_tok(ctx, T_UNION, &($3), $2);
			else
				$$ = C99Nalloc_tok(ctx, T_STRUCT, &($3), $2);
			free($1.str);
			c99c_loc_cpy_start(&$$->loc, &$1.loc);
		}
		| C_STRUCT attr_var_pre C_TYPENAME {
			if (*($1.str) == 'u')
				$$ = C99Nalloc(ctx, T_UNION, ctx->c_typename, $2);
			else
				$$ = C99Nalloc(ctx, T_STRUCT, ctx->c_typename, $2);
			C99Nloc_cpy_start(ctx, $$, $3);
			free($1.str);
			ctx->c_typename = NULL;
			C99Nfree(ctx, $3);
		}
		| str_head '{' '}'    { $$ = $1; C99Nchappend(ctx, $1, C99Nalloc_wloc(ctx, T_DECL, NULL, NULL, &ctx->loc_invalid, &ctx->loc_invalid)); }
		;

attr_var_pre:
		  /* empty */        { $$ = NULL; }
		| attr_spec_list     { $$ = $1; c99c_name_all_siblings(ctx, $$, "pre"); }
		;

attr_var_post:
		  /* empty */        { $$ = NULL; }
		| attr_spec_list     { $$ = $1; c99c_name_all_siblings(ctx, $$, "post"); }
		;

attr_spec_list:
		  attribute_specifier                    { $$ = $1; }
		| attr_spec_list attribute_specifier     { $$ = C99Nappend(ctx, $1, $2); }
		;

str_head:
		  C_STRUCT attr_var_pre {
			c99c_loc_t *eloc;
			if ($2 == NULL)
				eloc = &$1.loc;
			else
				eloc = &$2->loc;
			if (*($1.str) == 'u')
				$$ = C99Nalloc_wloc(ctx, T_UNION, NULL, $2, &$1.loc, eloc);
			else
				$$ = C99Nalloc_wloc(ctx, T_STRUCT, NULL, $2, &$1.loc, eloc);
			free($1.str);
		}
		| C_STRUCT attr_var_pre str_name {
			if (*($1.str) == 'u')
				$$ = C99Nalloc_tok(ctx, T_UNION, &($3), $2);
			else
				$$ = C99Nalloc_tok(ctx, T_STRUCT, &($3), $2);
			c99c_loc_cpy_start(&$$->loc, &$1.loc);
			free($1.str);
		}
		;

/* name of a structure */
str_name:
	  C_NAME        { $$ = $1; }
	| C_TYPENAME    { $$.str = c99c_strclone($1->name); c99c_loc_cpy_start(&($$.loc), &($1->loc)); C99Nfree(ctx, $1); free(ctx->c_typename); ctx->c_typename = NULL; }
	;

struct_dcl_list:
		  struct_declaration                    { if ($1->type == T_PRAGMA) $$ = $1; else $$ = C99Nalloc_wloc(ctx, T_DECL, NULL, $1, NULL, &$1->term_loc); }
		| struct_dcl_list struct_declaration    { if ($2->type == T_PRAGMA) $$ = C99Nappend(ctx, $1,  $2); else $$ = C99Nappend(ctx, $1, C99Nalloc_wloc(ctx, T_DECL, NULL, $2, NULL, &$2->term_loc)); }
		;

struct_declaration:
		  type_sq_list struct_declarator_list opt_extra_semi { c99c_node_t *l = C99Nlast(ctx, $1); $$ = C99Nalloc_wloc(ctx, T_TYPE, NULL, $1, &$1->loc, l->type == T_TYPECUSTOM ? &$2->loc : &$1->loc); C99Nappend(ctx, $$, $2); c99c_loc_cpy(&$$->term_loc, &$3.loc); }
		| preproc          { $$ = $1; }
		;

opt_extra_semi:
		  ';'                     { $$ = $1; }
		| opt_extra_semi ';'      { $$ = $2; c99c_werror(ctx, "extra ; in struct"); }
		;

type_sq_list:
		  type_sq                   { $$ = $1; }
		| type_sq_list type_sq      { $$ = C99Nappend(ctx, $1, $2);  }
		;

struct_declarator_list:
		  struct_declarator                                { $$ = $1; }
		| struct_declarator_list ',' struct_declarator     { $$ = C99Nappend(ctx, $1, $3); }
		;

struct_declarator:
		  declarator attr_var_post { $$ = $1; C99Nchappend(ctx, $$, $2); }
		| ':' be                   { $$ = C99Nalloc_wloc(ctx, T_ID, c99c_strclone("<anonymous>"), C99Nalloc_wloc(ctx, T_BITS, NULL, $2, &$1.loc, NULL), &ctx->loc_invalid, &ctx->loc_invalid); }
		| declarator ':' be        { $$ = $1; C99Nchappend(ctx, $$, C99Nalloc_wloc(ctx, T_BITS, NULL, $3, &$2.loc, NULL)); }
		| /* unnamed member */     { $$ = C99Nalloc_wloc(ctx, T_ID, c99c_strclone("<anonymous>"), NULL, &ctx->loc_invalid, &ctx->loc_invalid); }
		;

/* declarator with attributes; always preceeded by attributes */
xnfdeclarator:
		  declarator attr_var_post                       { $$ = $1; C99Nchappend(ctx, $$, $2); }
		| declarator C_ASM '(' string ')' attr_var_post  { $$ = $1; C99Nchappend(ctx, $$, C99Nalloc(ctx, T_ASM, NULL, $4)); C99Nchappend(ctx, $$, $6); }
		;

xnfdeclarator_typename:
		  declarator_typename attr_var_post                       { $$ = $1; C99Nchinsert(ctx, $$, $2); }
		| declarator_typename C_ASM '(' string ')' attr_var_post  { $$ = $1; C99Nchappend(ctx, $$, C99Nalloc(ctx, T_ASM, NULL, $4)); C99Nchinsert(ctx, $$, $6); }
		;

/*
 * Handles declarations and assignments.
 */
init_declarator:
		  xnfdeclarator                                    { $$ = $1; }
		| xnfdeclarator '=' ae                             { $$ = $1; C99Nchappend(ctx, $$, C99Nalloc_wloc(ctx, T_INIT, NULL, $3, &$2.loc, &$2.loc)); }
		| xnfdeclarator '=' '{' init_list optcomma '}'     { $$ = $1; C99Nchappend(ctx, $$, C99Nalloc_wloc(ctx, T_INIT, NULL, C99Nalloc_wloc(ctx, T_INITLIST, NULL, $4, &$3.loc, &$6.loc), &$2.loc, &$2.loc)); }
		| xnfdeclarator '=' '{' '}'                        { $$ = $1; C99Nchappend(ctx, $$, C99Nalloc_wloc(ctx, T_INIT, NULL, C99Nalloc_wloc(ctx, T_INITLIST, NULL, NULL, &$3.loc, &$4.loc), &$2.loc, &$2.loc)); }
		| xnfdeclarator '=' addrlbl                        { $$ = $1; C99Nchappend(ctx, $$, C99Nalloc_wloc(ctx, T_INIT, NULL, $3, &$2.loc, &$2.loc)); }
		;

init_declarator_typename:
		  xnfdeclarator_typename                                    { $$ = $1; }
		| xnfdeclarator_typename '=' ae                             { $$ = $1; C99Nchappend(ctx, $$, C99Nalloc_wloc(ctx, T_INIT, NULL, $3, &$2.loc, &$2.loc)); }
		| xnfdeclarator_typename '=' '{' init_list optcomma '}'     { $$ = $1; C99Nchappend(ctx, $$, C99Nalloc_wloc(ctx, T_INIT, NULL, C99Nalloc(ctx, T_INITLIST, NULL, $4), &$2.loc, &$2.loc)); }
		| xnfdeclarator_typename '=' '{' '}'                        { $$ = $1; C99Nchappend(ctx, $$, C99Nalloc_wloc(ctx, T_INIT, NULL, C99Nalloc(ctx, T_INITLIST, NULL, NULL), &$2.loc, &$2.loc)); }
		| xnfdeclarator_typename '=' addrlbl                        { $$ = $1; C99Nchappend(ctx, $$, C99Nalloc_wloc(ctx, T_INIT, NULL, $3, &$2.loc, &$2.loc)); }
		;


initializer:
		  ae %prec ','                   { $$ = $1; }
		| addrlbl                        { $$ = $1; }
		| '{' init_list optcomma '}'     { $$ = C99Nalloc_wloc(ctx, T_INITLIST, NULL, $2, &$1.loc, &$4.loc); }
		| '{' '}'                        { $$ = C99Nalloc_wloc(ctx, T_INITLIST, NULL, NULL, &$1.loc, &$2.loc); }
		;

/* initializer content */
init_list:
		  designation initializer                   {
			if ($1 == NULL) {
				/* empty desingation is for "1", "2" in "int a[] = {1, 2, };" */
				$$ = $2;
			}
			else {
				c99c_node_t *n;
				$$ = C99Nalloc(ctx, T_OP, c99c_strclone("= binary"), C99Nappend(ctx, $1, $2));
				for(n = $1; is_designator(n->children); n = n->children) ;
				C99Nchinsert(ctx, n, C99Nalloc_wloc(ctx, T_ID, c99c_strclone("<anonymous>"), NULL, &ctx->loc_invalid, &ctx->loc_invalid)); /* explicit anonymous */
			}
		}
		| init_list ','  designation initializer    {
			if ($3 != NULL) {
				c99c_node_t *n;
				/* empty desingation is for "1", "2" in "int a[] = {1, 2, };" */
				$$ = C99Nappend(ctx, $1, C99Nalloc(ctx, T_OP, c99c_strclone("= binary"), C99Nappend(ctx, $3, $4)));

				for(n = $3; is_designator(n->children); n = n->children) ;
				C99Nchinsert(ctx, n, C99Nalloc_wloc(ctx, T_ID, c99c_strclone("<anonymous>"), NULL, &ctx->loc_invalid, &ctx->loc_invalid)); /* explicit anonymous */
			}
			else
				$$ = C99Nappend(ctx, $1, $4);
		}
		;

designation:
		  designator_list '='           { $$ = $1; }
		| '[' be C_ELLIPSIS be ']' '='  { $$ = C99Nop(ctx, "[] binary", C99Nop(ctx, "... binary", $2, $4, &$3.loc, &$3.loc), NULL, &$1.loc, &$5.loc); }
		|                               { $$ = NULL; }
		;

designator_list:
		  designator                   { $$ = $1; }
		| designator_list designator   { $$ = $2; C99Nchinsert(ctx, $2, $1); }
		;

designator:
		  '[' be ']'         { $$ = C99Nalloc_wloc(ctx, T_OP, c99c_strclone("[] binary"), $2, &$1.loc, &$3.loc); }
		| C_STROP C_NAME     { $$ = C99Nalloc_tok(ctx, T_OP, &($1), C99Nalloc_tok(ctx, T_ID, &($2), NULL)); }
		| C_CUSTOP_B C_NAME  { $$ = C99Nalloc_tok(ctx, T_OP, &($1), C99Nalloc_tok(ctx, T_ID, &($2), NULL)); }
		;

optcomma:
		  /* VOID */
		| ','
		;

/*        STATEMENTS        */
compoundstmt:
		  beginc block_item_list endc      { $$ = $3; $$->children = $2; c99c_loc_cpy_end(&($$->loc), ctx->loc_current); }
		| beginc endc                      { $$ = $2; }
		;

beginc:
		'{'      { c99c_scope_enter(ctx); ctx->scope->compound = C99Nalloc(ctx, T_BLOCK, NULL, NULL); }
		;

endc:
		'}'      { $$ = ctx->scope->compound; c99c_loc_cpy_end(&$$->loc, &$1.loc); c99c_scope_leave(ctx); }
		;

begin:
		'{'      { c99c_scope_enter(ctx); }
		;

end:
		'}'      { c99c_scope_leave(ctx); }
		;

statement:
		  e ';'           { $$ = C99Nalloc_wloc(ctx, T_STATEMENT, c99c_strclone("expr"), $1, NULL, &$2.loc); }
		| compoundstmt    { $$ = $1; }
		| ifprefix virt_statement {
			$$ = $1;
			C99Nchappend(ctx, $$, C99Nalloc(ctx, T_THEN, NULL, $2));
			C99Nloc_cpy_end(ctx, $$, $2);
		}
		| ifelprefix virt_statement {
			$$ = $1;
			C99Nappend(ctx, $$->children, C99Nalloc(ctx, T_ELSE, NULL, $2));
			C99Nloc_cpy_end(ctx, $$, $2);
		}
		| whprefix virt_statement                       {
			$$ = $1;
			C99Nchappend(ctx, $$, $2);
			C99Nloc_cpy_end(ctx, $$, $2);
			}
		| doprefix virt_statement C_WHILE '(' e ')' ';' {
			$$ = $1;
			C99Nchappend(ctx, $$, $2);
			C99Nchappend(ctx, $$, C99Nalloc_tok(ctx, T_STATEMENT, &($3), C99Nalloc(ctx, T_CONDITION, NULL, $5)));
			c99c_loc_cpy_end(&$$->loc, &$7.loc);
		}
		| forprefix .e ')' virt_statement {
			$$ = $1;
			C99Nchappend(ctx, $$, C99Nalloc_wloc(ctx, T_CONDITION, c99c_strclone("iteration"), $2, C99NULL_LOC_OR_INV($2), C99NULL_LOC_OR_INV($2)));
			C99Nchappend(ctx, $$, $4);
			C99Nloc_cpy_end(ctx, $$, $4);
		}
		| switchpart statement    { $$ = $1; C99Nchappend(ctx, $$, $2); }
		| C_BREAK  ';'            { $$ = C99Nalloc_tok(ctx, T_STATEMENT, &($1), NULL);    c99c_loc_cpy_end(&$$->loc, &$2.loc);}
		| C_CONTINUE  ';'         { $$ = C99Nalloc_tok(ctx, T_STATEMENT, &($1), NULL);    c99c_loc_cpy_end(&$$->loc, &$2.loc);}
		| C_RETURN  ';'           { $$ = C99Nalloc_tok(ctx, T_STATEMENT, &($1), NULL);    c99c_loc_cpy_end(&$$->loc, &$2.loc);}
		| C_RETURN e  ';'         { $$ = C99Nalloc_tok(ctx, T_STATEMENT, &($1), $2);   c99c_loc_cpy_end(&$$->loc, &$3.loc); }
		| C_GOTO C_NAME ';'       { $$ = C99Nalloc_tok(ctx, T_STATEMENT, &($1), C99Nalloc_tok(ctx, T_ID, &($2), NULL));    c99c_loc_cpy_end(&$$->loc, &$3.loc);}
		| C_GOTO '*' e ';'        { $$ = C99Nalloc_tok(ctx, T_STATEMENT, &($1), C99Nop(ctx, "* unary", $3, NULL, &$2.loc, &$2.loc));   c99c_loc_cpy_end(&$$->loc, &$4.loc);}
		| asmstatement ';'        { $$ = C99Nalloc(ctx, T_ASM, NULL, $1);    c99c_loc_cpy_end(&$$->loc, &$2.loc);}
		| ';'                     { $$ = C99Nalloc_wloc(ctx, T_STATEMENT, NULL, NULL, &$1.loc, &$1.loc); }
		| error  ';'              { $$ = NULL; }
		| error '}'               { $$ = NULL; }
		| label statement         { $$ = $1; C99Nappend(ctx, $1, $2); }
		| switch_label statement  { $$ = $1; C99Nappend(ctx, $1, $2); }
		| preproc                 { $$ = $1; }
		;

/* wrap statement in a virtual block if it has a ->next */
virt_statement:
	statement   {
		if (($1 != NULL) && ($1->next != NULL))
			$$ = C99Nalloc(ctx, T_BLOCK, c99c_strclone("virtual"), $1);
		else
			$$ = $1;
		}
	;
asmstatement:
		  C_ASM asm_mvol '(' string ')'        { $$ = $4; }
		| C_ASM asm_mvol '(' string xasm ')'   { $$ = $4; C99Nchappend(ctx, $$, $5); }
		;

asm_mvol:
		  /* empty */      { $$ = NULL; }
		| C_QUALIFIER      { $$ = C99Nalloc_tok(ctx, T_QUALIFIER, &($1), NULL); }
		;

xasm:
		   ':' oplist                          { $$ = C99Nappend(ctx, C99Nalloc(ctx, T_STATEMENT, c99c_strclone(":"), NULL), $2); }
		|  ':' oplist ':' oplist               { $$ = C99Nappend(ctx, $2, C99Nappend(ctx, C99Nalloc(ctx, T_STATEMENT, c99c_strclone(":"), NULL), $4)); }
		|  ':' oplist ':' oplist ':' cnstr     { $$ = C99Nappend(ctx, $2, C99Nappend(ctx, C99Nalloc(ctx, T_STATEMENT, c99c_strclone(":"), NULL), $4)); C99Nappend(ctx, $$, C99Nappend(ctx, C99Nalloc(ctx, T_STATEMENT, c99c_strclone(":"), NULL), $6)); }
		;

/* asm oper list */
oplist:
		  /* nothing */    { $$ = NULL; }
		| oper             { $$ = $1; }
		;

/* asm oper */
oper:
		  string '(' e ')'             { $$ = $1; C99Nchappend(ctx, $$, $3); }
		| oper ',' string '(' e ')'    { $$ = C99Nappend(ctx, $1, C99Nappend(ctx, C99Nalloc(ctx, T_STATEMENT, c99c_strclone(","), NULL), $3)); C99Nchappend(ctx, $3, $5); }
		;

/* last part of an asm statement */
cnstr:
		  string           { $$ = $1; }
		| cnstr ',' string { $$ = C99Nappend(ctx, $1, $3); }
		;

label:
		  C_NAME ':'       { $$ = C99Nalloc_wloc(ctx, T_LABEL, c99c_strclone("goto"), C99Nalloc_tok(ctx, T_ID, &($1), NULL), NULL, &$2.loc); }
		| C_TYPENAME ':'   { $$ = C99Nalloc_wloc(ctx, T_LABEL, c99c_strclone("goto"), C99Nalloc(ctx, T_ID, ctx->c_typename, NULL), NULL, &$2.loc); C99Nloc_cpy_start(ctx, $$, $1); $$->children->loc.twin_parent = $1->loc.twin_parent; ctx->c_typename = NULL; C99Nfree(ctx, $1); }
		;

switch_label:
		  C_CASE be ':'                 { $$ = C99Nalloc_tok(ctx, T_LABEL, &($1), $2); c99c_loc_cpy_end(&$$->loc, &$3.loc); }
		| C_CASE be C_ELLIPSIS be ':'   { $$ = C99Nalloc_tok(ctx, T_LABEL, &($1), C99Nop(ctx, "... binary", $2, $4, &$3.loc, &$3.loc)); c99c_loc_cpy_end(&$$->loc, &$5.loc); }
		| C_DEFAULT ':'                 { $$ = C99Nalloc_tok(ctx, T_LABEL, &($1), NULL); c99c_loc_cpy_end(&$$->loc, &$2.loc); }
		;

doprefix:
		C_DO { $$ = C99Nalloc_tok(ctx, T_STATEMENT, &($1), NULL); }
		;

ifprefix:
		C_IF '(' e ')' { $$ = C99Nalloc_tok(ctx, T_STATEMENT, &($1), C99Nalloc(ctx, T_CONDITION, NULL, $3)); }
		;

ifelprefix:
		ifprefix statement C_ELSE {
			$$ = $1;
			C99Nchappend(ctx, $$, C99Nalloc(ctx, T_THEN, NULL, $2));
		}
		;

whprefix:
		C_WHILE  '('  e  ')'    { $$ = C99Nalloc_tok(ctx, T_STATEMENT, &($1), C99Nalloc(ctx, T_CONDITION, NULL, $3)); }
		;
forprefix:
		  C_FOR  '('  .e  ';' .e  ';' {
			$$ = C99Nalloc_tok(ctx, T_STATEMENT, &($1), NULL);
			C99Nchappend(ctx, $$, C99Nalloc_wloc(ctx, T_CONDITION, c99c_strclone("init"), $3, C99NULL_LOC_OR_INV($3), C99NULL_LOC_OR_INV($3)));
			C99Nchappend(ctx, $$, C99Nalloc_wloc(ctx, T_CONDITION, c99c_strclone("condition"), $5, C99NULL_LOC_OR_INV($5), C99NULL_LOC_OR_INV($5)));
		}
		| C_FOR '(' declaration .e ';' {
			$$ = C99Nalloc_tok(ctx, T_STATEMENT, &($1), NULL);
			C99Nchappend(ctx, $$, C99Nalloc(ctx, T_CONDITION, c99c_strclone("init"), $3));
			C99Nchappend(ctx, $$, C99Nalloc_wloc(ctx, T_CONDITION, c99c_strclone("condition"), $4, C99NULL_LOC_OR_INV($4), C99NULL_LOC_OR_INV($4)));
		}
		;

switchpart:
		C_SWITCH  '('  e ')' { $$ = C99Nalloc_tok(ctx, T_STATEMENT, &($1), C99Nalloc(ctx, T_CONDITION, NULL, $3)); }
		;


/*        EXPRESSIONS        */

/* expression or empty */
.e:
		  e    { $$ = $1; }
		|      { $$ = NULL; }
		;

/* list of expressions */
elist:
		  /* empty */            { $$ = NULL; }
		| ae %prec ','           { $$ = C99Nalloc(ctx, T_ARG, NULL, $1); }
		| elist  ','  ae         { $$ = C99Nappend(ctx, $1, C99Nalloc(ctx, T_ARG, NULL, $3)); }
		| elist  ','  cast_type  { /* hack for stdarg */ $$ = C99Nappend(ctx, $1, C99Nalloc(ctx, T_ARG, NULL, $3)); }
		;

/*
 * expression; precedence order of operators.
 */
e:
		  e ',' ae         {
			if ($1->type == T_EXPRLIST) {
				$$ = $1;
				C99Nchappend(ctx, $$, $3);
			}
			else
				$$ = C99Nalloc(ctx, T_EXPRLIST, NULL, C99Nappend(ctx, $1, $3));
		}
		| ae               { $$ = $1; }
		;
/* assignment-expression, it cannot be a comma separated list of expressions */
ae:
		  term '=' ae      { $$ = C99Nop(ctx, "= binary", $1, $3, &$2.loc, &$2.loc); }
		| term '=' addrlbl { $$ = C99Nop(ctx, "= binary", $1, $3, &$2.loc, &$2.loc); }
		| be               { $$ = $1; }
		| term C_ASOP ae   {
				$2.str = realloc($2.str, 64);
				strcat($2.str, " binary");
				$$ = C99Nop_tok(ctx, &($2), $1, $3);
				free($2.str);
			}

		;
/* binary-, ternary- or unary-expression, it cannot be an assignment */
be:
		  be '?' e ':' be  { $$ = C99Nop(ctx, "? trenary", $1, $3, &$2.loc, &$2.loc); C99Nchappend(ctx, $$, $5); }
		| be '?' ':' be    { $$ = C99Nop(ctx, "? trenary", $1, NULL, &$2.loc, &$2.loc); C99Nchappend(ctx, $$, $4); }
		| be C_OROR be     { $2.str = realloc($2.str, 64); strcat($2.str, " binary"); $$ = C99Nop_tok(ctx, &($2), $1, $3); free($2.str); }
		| be C_ANDAND be   { $2.str = realloc($2.str, 64); strcat($2.str, " binary"); $$ = C99Nop_tok(ctx, &($2), $1, $3); free($2.str); }
		| be '|' be        { $$ = C99Nop(ctx, "| binary", $1, $3, &$2.loc, &$2.loc); }
		| be '^' be        { $$ = C99Nop(ctx, "^ binary", $1, $3, &$2.loc, &$2.loc); }
		| be '&' be        { $$ = C99Nop(ctx, "& binary", $1, $3, &$2.loc, &$2.loc); }
		| be C_EQUOP be    { $2.str = realloc($2.str, 64); strcat($2.str, " binary"); $$ = C99Nop_tok(ctx, &($2), $1, $3); free($2.str); }
		| be C_RELOP be    { $2.str = realloc($2.str, 64); strcat($2.str, " binary"); $$ = C99Nop_tok(ctx, &($2), $1, $3); free($2.str); }
		| be C_SHIFTOP be  { $2.str = realloc($2.str, 64); strcat($2.str, " binary"); $$ = C99Nop_tok(ctx, &($2), $1, $3); free($2.str); }
		| be '+' be        { $$ = C99Nop(ctx, "+ binary", $1, $3, &$2.loc, &$2.loc); }
		| be '-' be        { $$ = C99Nop(ctx, "- binary", $1, $3, &$2.loc, &$2.loc); }
		| be C_DIVOP be    { $2.str = realloc($2.str, 64); strcat($2.str, " binary"); $$ = C99Nop_tok(ctx, &($2), $1, $3); free($2.str); }
		| be '*' be        { $$ = C99Nop(ctx, "* binary", $1, $3, &$2.loc, &$2.loc); }
		| be C_CUSTOP_B be { $$ = C99Nop(ctx, "custom binary", $1, $3, &$2.loc, &$2.loc); }
		| term             { $$ = $1; }
		;

/* jump address pointer, for gcc && goto trick */
addrlbl:
		C_ANDAND C_NAME    { $$ = C99Nop(ctx, "&& unary", C99Nalloc_tok(ctx, T_ID, &($2), NULL), NULL, &$1.loc, &$1.loc); free($1.str); }
		;

term:
		  term C_INCOP       { char *s; s = c99c_strclone("post?? unary"); memcpy(s+4, $2.str, 2); $$ = C99Nop(ctx, s, $1, NULL, &$2.loc, &$2.loc); free($2.str); free(s); }
		| '*' term           { $$ = C99Nop(ctx, "* unary", $2, NULL, &$1.loc, &$1.loc); }
		| '&' term           { $$ = C99Nop(ctx, "& unary", $2, NULL, &$1.loc, &$1.loc); }
		| '-' term           { $$ = C99Nop(ctx, "- unary", $2, NULL, &$1.loc, &$1.loc); }
		| '+' term           { $$ = C99Nop(ctx, "+ unary", $2, NULL, &$1.loc, &$1.loc); }
		| C_UNOP term        { $$ = C99Nop_tok(ctx, &($1), $2, NULL); free($1.str); }
		| C_INCOP term       { char *s; s = c99c_strclone("pre?? unary"); memcpy(s+3, $1.str, 2); $$ = C99Nop(ctx, s, $2, NULL, &$1.loc, &$1.loc); free($1.str); free(s); }
		| C_SIZEOF term      { $$ = C99Nop(ctx, "sizeof unary", $2, NULL, &$1.loc, &$1.loc); }
		| '(' cast_type ')' term  %prec C_INCOP              { $$ = C99Nop(ctx, "cast binary", C99Nalloc(ctx, T_TYPE, NULL, $2), $4, &$1.loc, &$3.loc); }
		| '(' cast_type ')' '{' init_list optcomma '}'       { $$ = C99Nop(ctx, "cast binary", C99Nalloc(ctx, T_TYPE, NULL, $2), C99Nalloc_wloc(ctx, T_INITLIST, NULL, $5, &$4.loc, &$7.loc), &$1.loc, &$3.loc); }
		| C_SIZEOF '(' cast_type ')'  %prec C_SIZEOF         { $$ = C99Nop(ctx, "sizeof unary", C99Nalloc(ctx, T_TYPE, NULL, $3), NULL, &$1.loc, &$1.loc); }
		| C_ALIGNOF '(' cast_type ')'                        { $$ = C99Nop(ctx, "alignof unary", $3, NULL, &$1.loc, &$1.loc); }
		| term '[' e ']'                        { $$ = C99Nalloc_wloc(ctx, T_OP, c99c_strclone("[] binary"), $1, &$2.loc, &$4.loc); C99Nchappend(ctx, $$, $3); }
		| C_NAME  '(' elist ')'                 { $$ = C99Nalloc_tok(ctx, T_CALL, &($1), C99Nappend(ctx, C99Nalloc_wloc(ctx, T_ID, c99c_strclone($1.str), NULL, &$1.loc, &$1.loc), C99Nalloc_wloc(ctx, T_ARGLIST, NULL, $3, &$2.loc, &$4.loc))); c99c_loc_cpy_end(&$$->loc, &$4.loc); }
		| term  '(' elist ')'                   { $$ = C99Nalloc(ctx, T_CALL, c99c_strclone(extract_id_from_cast(ctx, $1)), $1); C99Nchappend(ctx, $$, C99Nalloc_wloc(ctx, T_ARGLIST, NULL, $3, &$2.loc, &$4.loc)); }
		| term C_STROP C_NAME                   { $$ = C99Nop_tok(ctx, &($2), $1, C99Nalloc_tok(ctx, T_ID, &($3), NULL)); free($2.str); }
		| term C_CUSTOP_B C_NAME                { $$ = C99Nop_tok(ctx, &($2), $1, C99Nalloc_tok(ctx, T_ID, &($3), NULL)); free($2.str); }
		| term C_STROP C_TYPENAME               { $$ = C99Nop_tok(ctx, &($2), $1, C99Nalloc(ctx, T_ID, ctx->c_typename, NULL)); C99Nloc_cpy_start(ctx, $$, $3); ctx->c_typename = NULL; free($2.str); C99Nfree(ctx, $3); }
		| term C_CUSTOP_B C_TYPENAME            { $$ = C99Nop_tok(ctx, &($2), $1, C99Nalloc(ctx, T_ID, ctx->c_typename, NULL)); C99Nloc_cpy_start(ctx, $$, $3); ctx->c_typename = NULL; free($2.str); C99Nfree(ctx, $3); }
		| C_NAME %prec C_SIZEOF                 { $$ = C99Nalloc_tok(ctx, T_ID, &($1), NULL); }
		| C_ICON                                { $$ = C99Nalloc_tok(ctx, T_ICONST, &($1), NULL); }
		| C_FCON                                { $$ = C99Nalloc_tok(ctx, T_FCONST, &($1), NULL); }
		| string                                { $$ = $1; }
		| '('  e  ')'                           { $$ = C99Nalloc_wloc(ctx, T_EXPRBRACKET, NULL, $2, &$1.loc, &$3.loc); }
		| '(' begin block_item_list end ')'     { $$ = C99Nalloc_wloc(ctx, T_CSE, NULL, $3, &$1.loc, &$5.loc); }
		;

/* a single string or multiple strings appended */
string_:
		  C_STRING             { $$ = C99Nalloc_tok(ctx, T_SCONST, &($1), NULL); }
		| string_ C_STRING     { $$ = C99Nappend(ctx, $1, C99Nalloc_tok(ctx, T_SCONST, &($2), NULL)); }
		;

/* if we have multiple strings appended, wrap the list in SCONSTLIST */
string:
	string_                  { if ($1->next != NULL) $$ = C99Nalloc(ctx, T_SCONSTLIST, NULL, $1); else $$ = $1; }
	;

/* the inner part of casts, list of types */
cast_type:
		  type_sq_list                        { $$ = $1; }
		| type_sq_list abstract_declarator    { $$ = C99Nappend(ctx, $1, $2); }
		;

/* things that shouldn't have hit this stage:
   preprocessor directives the compiler is interested in */
preproc:
	  C_PRAGMA              { $$ = C99Nalloc_tok(ctx, T_PRAGMA,  &($1), NULL); }
	;

%%

static const char *extract_id_from_cast(c99tree_cc_t *ctx, c99c_node_t *n)
{
	if (n->type == T_EXPRBRACKET)
		n = n->children;
	switch(n->type) {
		case T_OP:
			if (strncmp(n->name, "cast ", 5) == 0)
				return n->children->next->name;
			if ((strcmp(n->name, "* unary") == 0) || (strcmp(n->name, "[] binary") == 0) ||
			    (strcmp(n->name, "-> binary") == 0) || (strcmp(n->name, ". binary") == 0))
				return "<anonymous>";
			{
				void *ctx_ = NULL;
				C99Nprint(ctx, &ctx_, " ? ", n, 0);
			}
			fprintf(stderr, "c99tree error: strange operator on function - please include that line in your bugreport!\n");
			abort();
		case T_ID:
			return "<anonymous>";
		default:
			{
				void *ctx_ = NULL;
				C99Nprint(ctx, &ctx_, " ? ", n, 0);
			}
			fprintf(stderr, "c99tree error: strange function cast %ld '%s' at %d - please include that line in your bugreport!\n", n->type, n->name, n->loc.sline);
			abort();
	}
}

static int is_designator(c99c_node_t *n)
{
	return (n->type == T_OP) && ((strcmp(n->name, "[] binary") == 0) || (strcmp(n->name, ". binary") == 0) || (strcmp(n->name, "-> binary") == 0));
}
