
# c99tree plugin

wrap_stmt()
{
	awk -v "prefix=$2" -v "suffix=$3" '
		($1 == "STATEMENT") {
			l1 = $NF
			gsub("[()]", "", l1)
			l2 = l1
			sub(":.*", "", l1)
			sub(".*:", "", l2)
			if (prefix != "")
				print l1,    prefix
			if (suffix != "")
				print l2+1,  suffix
		}
	' < "$4" | sort -n | sed "s/^/insert /" | $LIBDIR/c99tabedit "$1"
}

data_dump_AST()
{
	local TREECC tpaste tonly tparm src

	TREECC="$1"
	tpaste="$2"
	tonly="$3"
	tparm="$4"
	src="$5"

	eval $TREECC \
	--gtx "enum=   [] /+ e=[type == ENUM] /{1} [type == ID]" \
	--gtx "unistr= ([] /+ s=[type == STRUCT || type == UNION] /{1} d=[type == DECL] /{1} i=[type==ID]) && t=(d /{1} [type==TYPE])" \
	"-L%o:%O" -Tawk --paste "$tpaste" $tonly $tparm "$src"
}

awk_dump_AST()
{
	local TREECC tpaste tonly tparm src

	TREECC="$1"
	tpaste="$2"
	tonly="$3"
	tparm="$4"
	src="$5"

	eval $TREECC "-L%o:%O" -Tawk --paste "$tpaste" $tonly $tparm "$src"

}

cmd_edit()
{
	local cmd src tparm pn tree outfn CC ccparm tpaste tonly noncomp

	cmd="$1"
	shift 1

	CC=/usr/bin/cc

	while test $# -gt 0
	do
		case "$1" in
			-o) shift 1; outfn=$1 ;;
			--CC) shift 1; CC=$1 ;;
			--keep-cpp) tonly="$tonly $1";;
			--paste) shift 1; tpaste="'$1'";;
			-print-sysroot) noncomp="$noncomp $1" ;;
			-I|-D|--gtx) tparm="$tparm $1 '$2'"; shift 1;;
			-I*|-i|-D*|-U*|-L*) tparm="$tparm '$1'";;
			-W*|-l*|-g*|-f*|-m*|-pedantic|-ansi|-O*|-p*|-std*|-E|-S|-c|--no-builtin|-f*|-MD) ccparm="$ccparm $1";;
			--awk-s) awk_script=$2; shift 1;;
			--awk-f) awk_files="$awk_files -f '$2'"; shift 1;;
			-w) shift 1; wparm="$wparm -v $1" ;;
			-x) shift 1; ccparm="$ccparm -x $1";;
			-*)
				pn=${1##--}
				pn=${pn##-}
				shift 1
				export edit_arg_$pn="$1"
				;;
			*)
				if test -z "$src"
				then
					src="$1"
				else
					echo "Multiple source files not supported ($src, $1)" >&2
					exit 1
				fi
		esac
		shift 1
	done

	case "$cmd" in
		print_fn)
			echo "src $src"
			echo "out $outfn"
			;;
		print_fn_src)
			echo "$src"
			;;
		print_fn_out)
			echo "$out"
			;;
		print_ccparm)
			echo "$ccparm"
			;;
		print_cc)
			echo "$CC"
			;;
		print_cc_cmdline)
			if test -z "$outfn"
			then
				echo "$tparm $noncomp $ccparm $src"
			else
				echo "$tparm $noncomp $ccparm $src -o $outfn"
			fi
			;;
		print_noncomp)
			echo "$noncomp"
			;;
		wrap_stmt)
			tree=`tempfile`
			cout=`tempfile  --suffix .c`
			eval $TREECC -L%o:%O --paste "$tpaste" $tonly $tparm "$src" > $tree
			wrap_stmt "$src" "$edit_arg_stmtprefix" "$edit_arg_stmtsuffix" "$tree" > $cout
			if test ! -z "$outfn"
			then
				$CC $tparm $noncomp $ccparm -o "$outfn" $cout
			else
				cat $cout
			fi
			rm $cout
			rm $tree
			;;
		data-dump)
			tree=`tempfile`
			data_dump_AST "$TREECC" "$tpaste" "$tonly" "$tparm" "$src" > $tree
			awk -f $tree -f $LIBDIR/tree.awk -f $LIBDIR/gtx.awk -f $LIBDIR/map_data.awk -f $LIBDIR/map_data_cli.awk
			rm $tree
			;;
		data-edit)
			tree=`tempfile`
			data_dump_AST "$TREECC" "$tpaste" "$tonly" "$tparm" "$src" > $tree
			awk $wparm -f $tree -f $LIBDIR/tree.awk -f $LIBDIR/gtx.awk -f $LIBDIR/map_data.awk -f $LIBDIR/map_data_edit.awk | cmd_tabedit "$src"
			rm $tree
			;;
		awk)
			tree=`tempfile`
			awk_dump_AST "$TREECC" "$tpaste" "$tonly" "$tparm" "$src" > $tree
			if test ! -z "$awk_script"
			then
				tmp_awk=`tempfile  --suffix .awk`
				echo "$awk_script" > $tmp_awk
				awk_files="$awk_files -f $tmp_awk"
			fi
			awk $wparm -f $tree -f $LIBDIR/tree.awk -f $LIBDIR/gtx.awk -f $LIBDIR/map_data.awk $awk_files
			rm $tree
			if test ! -z "$awk_script"
			then
				rm $tmp_awk
			fi
			;;
	esac
}

cmd_tabedit()
{
	$LIBDIR/c99tabedit "$@"
}

help_edit()
{
	echo "Edit (and compile) a c source: complex manipulation"
	echo "Usage:"
	echo "c99tree edit [ecmd] [-o output] [--CC cc] [cc-params] [cmd-params] source"
	echo ""
	echo "Generic params:"
	echo " -o output          run CC on the edited source with -o output ^1"
	echo " --CC cc            use replace the default cc (/usr/bin/cc) with cc"
	echo ""
	echo "cc-params: usual parameters of a C compiler:"
	echo " -W, -l, -g, -f, -m, -pedantic, ansi, O, -p, -std,"
	echo " -E, -S, -c, -I, -i, -D, -U"
	echo ""
	echo "ecmd is the edit command; following is a list of ecmds with their cmd-params"
	echo " wrap_stmt          wrap each statement"
	echo "   --stmtprefix s   insert string s before each statement"
	echo "   --stmtsuffix s   insert string s after each statement"
	echo ""
	echo " print_fn           print the source and output file name"
	echo ""
	echo "Footnotes:"
	echo " ^1: without -o the edited C source is printed on stdout. With -o, the"
	echo "     command line can be used as a C compiler (e.g. CC in a Makefile)."
	echo "     This is also called the \"transparent mode\"."
	echo ""
}

help_tabedit()
{
		echo "Edit a c source: simple stream edit"
		echo "Usage:"
		echo "c99tree tabedit [-r] src"
		echo "Flags:"
		echo"  -r      raw text: do not interpret backslash characters in txt fields"
		echo ""
		echo "Edit commands are read from stdin, the C source is read from file src"
		echo "and the output is written to stdout. Edit commands need to be a list"
		echo "of commands sorted by start their offset. The list must be free of"
		echo "overlaps (in case of commands with both start and end offsets)."
		echo ""
		echo "Each command is in a new line, fields of the commands are separated by"
		echo "exactly one whitespace character."
		echo ""
		echo "Commands:"
		echo " insert soffs txt        - insert txt at soffs"
		echo " delete soffs eoffs      - delete everything between soffs and eoffs"
		echo " replace soffs eoffs txt - replace everything between soffs and eoffs with txt"
		echo ""
		echo "Offsets are byte offsets into the C source, counted from 0."
		echo ""
}
